//-----------------------------------------------------------------------------
// File: dxescape.cpp
//
// Desc: Demonstrates use of the Adaptoid's DirectInput Driver Escapes
//
// Copyright (c) 2000 Wish Technologies, Inc.
//-----------------------------------------------------------------------------

#include <windows.h>
#include <dinput.h>
#include "resource.h"
#include "dxsample.h"

//
//  *** Description of DirectInput Escapes supported by the Adaptoid driver
//
//
//  Many features of the Adaptoid can be accessed via the Driver Escape mechanism
//  provided by DirectInput.  Driver escapes can be accessed via the "Escape"
//  function of the IDirectInputDevice2 interface.
//
//  In order for a driver escape to work, you must have set the cooperative level
//  of the IDirectInput2 device to DISCL_EXCLUSIVE, and the joystick device must
//  be in the Aquired state.
//
//  The following escape commands are provided:
//
//    0x7834BB00:  Query API - pass in command #, returns 0xB0CAB0CA if supported
//    0x7834BB04:  Get device number (dword, 0 based sequential, ordered by port #)
//    0x7834BB05:  Get port name (text string which identifies USB port uniquely)
//    0x7834BB08:  Send command to rumble pack (DWORD 0=stop, 1=go)
//    0x7834BB0C:  Initialize pak (returns pak status bit flags)
//    0x7834BB0D:  Read from pak  (reads 32 bytes of data)
//    0x7834BB0E:  Write to pak   (writes 32 bytes of data)
//    0x7834BB10:  Set (or get) joystick clipping range parameter  (default 75)
//    0x7834BB11:  Set (or get) joystick corner stretch parameter  (default 10)
//    0x7834BB28:  Send command directly to controller - synchronous
//
//  Note that any parameters set by using the Driver Escapes will be reset to their
//  default values when the IDirectInputDevice2 interface is destroyed.
//


//------------------------------
//
// 0x7834BB00 - Query API
// 
// Inputs : DWORD (4 bytes), command number you are asking about
// Outputs: DWORD (4 bytes), either 0xBOCABOCA if supported, or 0xDEADBEEF if not
//
//
// The escape command 0x7834BB00 is the "query api" command the Adaptoid driver
// uses to let you know if a particular command is supported.  Pass the command
// you want to ask about as the input buffer, and the driver will return either
// 0xBOCABOCA if the command is supported or 0xDEADBEEF if it is not.
//
// To see if you are talking to the Adaptoid driver, use the "query api" escape
// command to ask if the "query api" command is supported.  You should get a
// 0xBOCABOCA response.  Any other response means you are not talking the
// Adaptoid driver.
//
// This method is more reliable than looking for certain strings or Guids in the
// DIDEVICEINSTANCE structure since the values returned in the DIDEVICEINSTANCE
// structure seem to change in each new version of DirectX.
//

BOOL IsAdaptoidCommandSupported( DWORD cmd)
{
    DIEFFESCAPE esc;
    DWORD inbuf, outbuf;
    HRESULT hr;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB00;   // command to determine if a command is supported
    esc.lpvInBuffer = &inbuf;
    esc.cbInBuffer = 4;
    esc.lpvOutBuffer = &outbuf;   
    esc.cbOutBuffer = 4;
    inbuf = cmd;                  // command that we are asking is supported
    outbuf = 0;                   

    hr = g_pJoystick->Escape(&esc);
    if (!FAILED(hr) && esc.cbOutBuffer == 4 && outbuf == 0xB0CAB0CA)
    {
        return TRUE;
    }

    return FALSE;
}



//--------------------------------------
//
// 0x7834BB04 - Get Device Number
//
// Inputs : none
// Outputs: DWORD (4 bytes), device number, 0-based sequential
// 
// 
// This command returns an integer which indicates which Adaptoid you are talking
// to.  When multiple Adaptoids are present, this device number will be in the
// same order as the USB port numbers the Adaptoids are plugged into.  This device
// order corresponds to the order used by the Adaptoid's Wish_NA1 driver interface.
//
// This device number will stay the same even after your computer is restarted.
// But please note that it may change at any time if another Adaptoid device is
// plugged into or unplugged from a lower USB port number.  For a more permanent
// device identifier, see the "Port Name" command below.
//
// This device number has no relation whatsoever to the DirectInput joystick number
// or control panel slot number, since the mapping of those indexes to actual devices
// potentially changes every time you reboot your system or power cycle a USB hub.
//

BOOL IsDeviceNumber()
{
    return IsAdaptoidCommandSupported( 0x7834BB04);
}

HRESULT GetDeviceNumber(DWORD *val)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB04;   // get device number
    esc.lpvInBuffer = NULL;
    esc.cbInBuffer = 0;
    esc.lpvOutBuffer = val;   
    esc.cbOutBuffer = 4;

    return g_pJoystick->Escape(&esc);
}



//--------------------------------------
//
// 0x7834BB05 - Get Port Name
// 
// Inputs : none
// Outputs: char string, null terminated - The name of the USB port the Adaptoid is in
//
// 
// This command returns a character string which describes the USB port that the
// Adaptoids is plugged into.  
//
// An example of a port name is "A24" which indicates that the Adaptoid is plugged
// into port 4 of an external hub which is plugged into port 2 of the root hub
// controlled by the first host controller (A) on the computer.
//
// The length of the string is 1+N where N is the depth of nested USB hubs. The USB
// spec indicates a maximum depth of 7 hubs, so the string should always be 8 chars
// or less.
//

BOOL IsPortName()
{
    return IsAdaptoidCommandSupported( 0x7834BB05);
}

HRESULT GetPortName(char *buf, int bufsize)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB05;   // get device number
    esc.lpvInBuffer = NULL;
    esc.cbInBuffer = 0;
    esc.lpvOutBuffer = buf;   
    esc.cbOutBuffer = bufsize;

    return g_pJoystick->Escape(&esc);
}



//--------------------------------------
//
// 0x7834BB08 - Send Rumble Command
// 
// Inputs : DWORD (4 bytes), value being 1=go, 0=stop
// Outputs: none
//
//
// This command either starts or stops the rumble pack rumbling.  Send a 1 to
// start rumbling or a 0 to stop rumbling.  
//
// This command returns immediately, without waiting for the controller or
// rumble pack to respond. All the comunication with the rumble pack is handled
// asyncronously by the kernel driver and/or the Adaptoid firmware. This method
// is faster and more efficient than using the Pak I/O or Direct I/O commands.
//
// Rumbling will stop automatically after approximately 4 seconds.  Rumbling 
// will also stop if the rumble pak is removed and reinserted.  To rumble 
// continuously, keep sending "1" commands.  The 4-second timeout is a failsafe
// to prevent continuous rumbling in the event a program crashes or forgets to
// stop the rumble pack.
//
// The rumble speed can be moderated by rapidly starting and stopping the rumble
// with varying duty cycles.
//

BOOL IsRumbleCommand()
{
    return IsAdaptoidCommandSupported( 0x7834BB08);
}

HRESULT SendRumbleCommand(DWORD cmd)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB08;   // send rumble command
    esc.lpvInBuffer = &cmd;  // 1=go, 0=stop
    esc.cbInBuffer = 4;
    esc.lpvOutBuffer = NULL;   
    esc.cbOutBuffer = 0;

    return g_pJoystick->Escape(&esc);
}



//--------------------------------------
//
// 0x7834BB0C - Initialize Pak
//
// Inputs : none
// Outputs: BYTE (1 byte), bitmask containing Pak status flags
//
// 0x7834BB0D - Read Pak
//
// Inputs : DWORD (4 bytes), starting Pak address to read from
// Outputs: BYTE array (32 bytes), the data that was read from the Pak
//
// 0x7834BB0E - Write Pak
//
// Inputs : DWORD & BYTE array (36 bytes total), starting Pak address & data to write
// Outputs: none
//
//
// These commands are used to communicate with whatever device is plugged into
// the "Pak" connector on the bottom of the controller.  
//
// The Pak appears as a 64K address space.  Reads or writes must be 32 bytes
// in size and must be on an even 32 byte boundary.  The 64K address space may
// be composed of RAM, ROM, or special purpose registers, depending on the
// type of Pak device.
//
// If a Pak has been inserted after the controller was powered up, you must
// first initialize the Pak interface before you can read or write to it.
// Reading or writing to an uninitialized Pak interface will return an error code.
// The Pak interface is automatically initialized when the controller is powered
// up.  When a Pak is removed, the Pak interface becomes uninitialized.  If you
// try to initialize the Pak interface when no Pak is present, it will have no
// effect.
//
// The fact that reads & writes are prevented until the pak interface is initialized
// is useful in the case where the user swaps mempaks just before a pak write is done.
// The pak write will fail before it has a chance to overwrite the contents of the new
// mempak, signaling you to examine the pak's contents first to see what you are
// writing to.
//
// When performing Pak I/O, the calling thread will block until the controller responds.
// This may take up to 15ms.  For more efficient rumble pak interaction, use the Rumble
// I/O escape.
//
// All Pak communication is protected from errors by a CRC.  If the CRC on a
// read or write doesn't match, an error code will be returned.  You should
// retry the read or write when a CRC error occurs.
//
// The BYTE returned from the Initialize Pak command consists of the following
// bits:
//
// 0x01:  Pak is present
// 0x02:  Pak interface was uninitialized before the call
// 0x04:  Address of last Pak I/O was invalid
//

BOOL IsPakIO()
{
    return IsAdaptoidCommandSupported( 0x7834BB0C) && 
           IsAdaptoidCommandSupported( 0x7834BB0D) &&
           IsAdaptoidCommandSupported( 0x7834BB0E);
}

HRESULT InitializePak(BYTE *status)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB0C;   // Initialize Pak
    esc.lpvInBuffer = NULL;
    esc.cbInBuffer = 0;
    esc.lpvOutBuffer = status;   
    esc.cbOutBuffer = 1;

    return g_pJoystick->Escape(&esc);
}

HRESULT ReadPak(DWORD addr, BYTE *data)
{
    DIEFFESCAPE esc;
    HRESULT hr;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB0D;   // Read 32 bytes from pak
    esc.lpvInBuffer = &addr;
    esc.cbInBuffer = 4;
    esc.lpvOutBuffer = data;   
    esc.cbOutBuffer = 32;

    hr = g_pJoystick->Escape(&esc);

    return hr;
}

HRESULT WritePak(DWORD addr, BYTE *data)
{
    DIEFFESCAPE esc;
    struct
    {
        DWORD addr;
        BYTE data[32];
    } buf;

    buf.addr = addr;
    memcpy(buf.data, data, 32);

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB0E;   // Write 32 bytes to pak
    esc.lpvInBuffer = &buf;
    esc.cbInBuffer = 36;
    esc.lpvOutBuffer = NULL;   
    esc.cbOutBuffer = 0;

    return g_pJoystick->Escape(&esc);
}



//--------------------------------------
//
// 0x7834BB10 - Range Clipping
// 
// Inputs : DWORD (4 bytes), optional, value to set the Range Clipping to
// Outputs: DWORD (4 bytes), optional, returns old value of Range Clipping parameter
// 
// 
// Range Clipping is used by the Adaptoid driver to make the physical range of
// the analog stick correspond to the logical range reported by DirectInput.
//
// The controller reports the position of the analog stick as a value from -128
// to +127 for each axis.  However, the stick is physically limited to a range
// of approximately -80 to +80 on each axis (+/- 4).
//
// Games on the PC expect the stick to be able to reach all positions in the
// logical range, so some sort of range transformation is needed.  The "Range
// Clipping" parameter is the position which is reported as full deflection.
// Futher deflection is reported as full deflection, effectively "clipping"
// away the extreme limits.  The default Range Clipping value is 75.
//
// If you wish to turn off this clipping and have access to the full range of
// values reported by the controller, set it to 128.  This will restore the
// original range reported by the controller, were full deflection of the stick
// is only about 65% of the logical range.
//

BOOL IsRangeClipping()
{
    return IsAdaptoidCommandSupported( 0x7834BB10);
}

HRESULT GetRangeClipping(DWORD *val)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB10;   // get/set range clipping
    esc.lpvInBuffer = NULL;
    esc.cbInBuffer = 0;
    esc.lpvOutBuffer = val;   
    esc.cbOutBuffer = 4;

    return g_pJoystick->Escape(&esc);
}

HRESULT SetRangeClipping( DWORD val)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB10;   // get/set range clipping
    esc.lpvInBuffer = &val;
    esc.cbInBuffer = 4;
    esc.lpvOutBuffer = NULL;   
    esc.cbOutBuffer = 0;

    return g_pJoystick->Escape(&esc);
}



//--------------------------------------
//
// 0x7834BB11 - Corner Stretch
// 
// Inputs : DWORD (4 bytes), optional, value to set the Corner Stretching to
// Outputs: DWORD (4 bytes), optional, returns old value of Corner Stretching parameter
// 
//
// Corner stretching is used by the Adaptoid driver to map the octagonal range of the
// analog stick to the square range expected by PC games.
//
// The analog stick has a physical range of approximately -80 to +80 on each axis, 
// but if you push it into a diagonal corner the maximum deflection is only about
// 66 on each axis.  This makes it impossible to produce a full diagonal deflection.
//
// This problem could be solved by setting the range clipping down to 65, but this
// would waste a lot of the horizonal and vertical range, losing accuracy and making
// the control seem over-responsive in those directions.
//
// A better solution is to "warp" the rectangular coordinate grid in the corners to
// allow full diagonal deflection while keeping the full horizonal and vertical range.
// The "Corner Stretch" parameter indicates how much to warp the grid.  The default
// value of 10 warps the maximum deflection of 65 out to 75 in the corners, matching
// the default range clipping of 75 in the horizonal and vertical directions.
//
// Note that the "warping" is smooth and linear across the full range.  It doesn't
// just "pop" you into the corner when you get near it.
//
// If you want to disable this stretching of the corners, set the Corner Stretch
// parameter to 0.
//

BOOL IsCornerStretch()
{
    return IsAdaptoidCommandSupported( 0x7834BB11);
}

HRESULT GetCornerStretch(DWORD *val)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB11;   // get/set corner stretch
    esc.lpvInBuffer = NULL;
    esc.cbInBuffer = 0;
    esc.lpvOutBuffer = val;   
    esc.cbOutBuffer = 4;

    return g_pJoystick->Escape(&esc);
}

HRESULT SetCornerStretch( DWORD val)
{
    DIEFFESCAPE esc;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB11;    // get/set corner stretch
    esc.lpvInBuffer = &val;
    esc.cbInBuffer = 4;
    esc.lpvOutBuffer = NULL;   
    esc.cbOutBuffer = 0;

    return g_pJoystick->Escape(&esc);
}


//--------------------------------------
//
// 0x7834BB28 - Send Command Directly to Controller
//
// Inputs : BYTE array, data to send to controller
// Outputs: BYTE array, data returned by controller
//
//
// This command sends data directly to the controller, waits for the controller
// to respond, and returns whatever data the controller returned.  This command
// may take up to 15ms to execute.
//
// The size of the input buffer is the number of bytes to send, and the size
// of the return buffer is the number of bytes requested back from the controller.
//
// You can send up to 4 bytes of data to the controller and receive up to 40
// bytes of data back from the controller, or you can send up to 40 bytes of
// data to the controller and receive up to 4 bytes back.
//
// If the controller does not respond, the return buffer size will be set to 0.
//
// This is a low level interface used to communicate directly with the controller.
// The details and format of the controller commands are not documented.
//

BOOL IsDirectCommand()
{
    return IsAdaptoidCommandSupported( 0x7834BB28);
}

HRESULT DirectCommand(BYTE *command, int commandsize, BYTE *result, int *resultsize)
{
    DIEFFESCAPE esc;
    HRESULT hr;

    esc.dwSize = sizeof(esc);
    esc.dwCommand = 0x7834BB28;   // Send direct controller command
    esc.lpvInBuffer = command;
    esc.cbInBuffer = commandsize;
    esc.lpvOutBuffer = result;   
    esc.cbOutBuffer = *resultsize;

    hr = g_pJoystick->Escape(&esc);

    if (SUCCEEDED(hr))
        *resultsize = esc.cbOutBuffer;

    return hr;
}


