//-----------------------------------------------------------------------------
// File: dxstuff.cpp
//
// Desc: DirectX specific stuff for DirectInput driver escapes sample
//
//-----------------------------------------------------------------------------

#define INITGUID
#include <windows.h>
#include <dinput.h>
#include "dxsample.h"
#include "resource.h"

//-----------------------------------------------------------------------------
// Global variables 
//-----------------------------------------------------------------------------
IDirectInput*           g_pDI       = NULL;         
IDirectInputDevice2*    g_pJoystick = NULL;
IDirectInputEffect*     g_pEffect   = NULL;
HINSTANCE               g_hInst     = NULL;
int                     g_nDevices  = 0;
int                     g_nButtonCount;
IDirectInputDevice2*    g_aDeviceList[16];
char                    g_pDeviceNames[16][MAX_PATH];
GUID                    g_gDeviceGuids[16];
int                     g_nCurrentDevice = -1;
BOOL                    g_bAcquired  = FALSE;
LONG                    g_xMin;
LONG                    g_xMax;
LONG                    g_yMin;
LONG                    g_yMax; 

//-----------------------------------------------------------------------------
// Local function-prototypes
//-----------------------------------------------------------------------------
void DestroyDevices(void);

BOOL CALLBACK EnumFFJoysticksCallback( LPCDIDEVICEINSTANCE pInst, 
                                       LPVOID lpvContext );




//-----------------------------------------------------------------------------
// Function: InitDirectInput
//
// Description: 
//      Get the DirectInput interface
//
//-----------------------------------------------------------------------------
HRESULT InitDirectInput( HWND hDlg )
{
    HRESULT hr;

    // Register with the DirectInput subsystem and get a pointer
    // to a IDirectInput interface we can use.
    hr = DirectInputCreate( g_hInst, DIRECTINPUT_VERSION, &g_pDI, NULL );
    if ( FAILED(hr) ) 
        return hr;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Function: EnumerateJoysticks
//
// Description: 
//      Create a list of joysticks which support the IDirectInput2 interface
//
//-----------------------------------------------------------------------------
HRESULT EnumerateJoysticks( HWND hDlg )
{
    HRESULT hr;

    // Out with the old devices
    DestroyDevices();

    // look for a joystick we can use for this sample program.
    hr = g_pDI->EnumDevices( DIDEVTYPE_JOYSTICK,
                             EnumFFJoysticksCallback,
                             NULL,
                             DIEDFL_ATTACHEDONLY );
    return hr;
}

//-----------------------------------------------------------------------------
// Function: EnumFFJoysticksCallback
//
// Description: 
//      Called once for each enumerated force feedback joystick.
//      If we find one, create a device interface on it so we can
//      play with it.
//
//-----------------------------------------------------------------------------
BOOL CALLBACK EnumFFJoysticksCallback( LPCDIDEVICEINSTANCE pInst, 
                                       LPVOID lpvContext )
{
    HRESULT                 hr;
    LPDIRECTINPUTDEVICE     pDevice;
    LPDIRECTINPUTDEVICE2    pDevice2;

    // obtain an interface to the enumerated force feedback joystick.
    hr = g_pDI->CreateDevice( pInst->guidInstance, &pDevice, NULL );

    // if it failed, then we can't use this joystick for some
    // bizarre reason.  (Maybe the user unplugged it while we
    // were in the middle of enumerating it.)  So continue enumerating
    if ( FAILED(hr) ) 
        return DIENUM_CONTINUE;

    // query for IDirectInputDevice2
    hr = pDevice->QueryInterface( IID_IDirectInputDevice2, (LPVOID *)&pDevice2);
    pDevice->Release(); // done with old interface now

    if ( FAILED(hr) ) 
    {
        return DIENUM_CONTINUE;  // try again
    }

    // we successfully created an IDirectInputDevice2.  
    g_aDeviceList[g_nDevices] = pDevice2;
    wsprintf( g_pDeviceNames[g_nDevices], "%s  /  %s", pInst->tszInstanceName, pInst->tszProductName);
    g_gDeviceGuids[g_nDevices] = pInst->guidInstance;
    g_nDevices++;

    return DIENUM_CONTINUE;
}



//-----------------------------------------------------------------------------
// Function: DestroyDevices
//
// Description: 
//      Destroy all the IDirectInput2 interfaces we created when enumerating
//      the joysticks.
//
//-----------------------------------------------------------------------------
void DestroyDevices()
{
    int i;

    for (i=0; i < g_nDevices; i++)
    {
        g_aDeviceList[i]->Unacquire();
        g_aDeviceList[i]->Release();
        g_aDeviceList[i] = NULL;
    }

    g_nDevices = 0;
    g_nCurrentDevice = -1;
    g_pJoystick = NULL;
}



//-----------------------------------------------------------------------------
// Function: FreeDirectInput
//
// Description: 
//      Free up the DirectInput interface (and anything else we created)
//
//-----------------------------------------------------------------------------
HRESULT FreeDirectInput()
{
    // Release any DirectInputEffect objects.
    if (NULL != g_pEffect) 
    {
        g_pEffect->Release();
        g_pEffect = NULL;
    }

    // Unacquire and release any DirectInputDevice objects.
    DestroyDevices();

    // Release any DirectInput objects.
    if (NULL != g_pDI) 
    {
        g_pDI->Release();
        g_pDI = NULL;
    }

    return S_OK;
}



//-----------------------------------------------------------------------------
// Function: UseJoystick
//
// Description: 
//      Set up and Acquire a specific joystick
//
//-----------------------------------------------------------------------------
HRESULT UseJoystick( HWND hDlg, int n)
{
    HRESULT hr;

    ReleaseJoystick( hDlg);
    
    if ( n >= g_nDevices || g_aDeviceList[n] == NULL)
    {
        return E_FAIL;
    }

    g_pJoystick = g_aDeviceList[n];

    // Set the data format to "simple joystick" - a predefined data format 
    //
    // A data format specifies which controls on a device we
    // are interested in, and how they should be reported.
    //
    // This tells DirectInput that we will be passing a
    // DIJOYSTATE structure to IDirectInputDevice2::GetDeviceState.
    //
    hr = g_pJoystick->SetDataFormat( &c_dfDIJoystick );
    if ( FAILED(hr) ) 
    {
        g_pJoystick = NULL;
        return hr;
    }


    // Find out how many buttons the joystick has
    DIDEVCAPS devcaps;
    devcaps.dwSize = sizeof(devcaps);
    hr = g_pJoystick->GetCapabilities( &devcaps );
    if ( FAILED(hr) ) 
    {
        g_pJoystick = NULL;
        return hr;
    }
    g_nButtonCount = devcaps.dwButtons;

    if (g_nButtonCount > 32)
        g_nButtonCount = 32;     // Limit button count to 32 since we're 
                               //  using DIJOYSTATE instead of DIJOYSTATE2


    // Get the axis ranges for the device. 
    DIPROPRANGE dipr;
    dipr.diph.dwSize       = sizeof(DIPROPRANGE);
	dipr.diph.dwHeaderSize = sizeof(dipr.diph);
	dipr.diph.dwHow        = DIPH_BYOFFSET;
    
	// Get the x-axis range property
    dipr.diph.dwObj        = DIJOFS_X;
    hr = g_pJoystick->GetProperty( DIPROP_RANGE, &dipr.diph );
    if( FAILED(hr) )
    {
        g_pJoystick = NULL;
        return hr;
    }

	g_xMin = dipr.lMin;
	g_xMax = dipr.lMax;

	// Get the y-axis range property
    dipr.diph.dwObj = DIJOFS_Y;
    hr = g_pJoystick->GetProperty( DIPROP_RANGE, &dipr.diph );
    if( FAILED(hr) )
    {
        g_pJoystick = NULL;
        return hr;
    }

	g_yMin = dipr.lMin;
	g_yMax = dipr.lMax;


    // Set the cooperativity level to let DirectInput know how
    // this device should interact with the system and with other
    // DirectInput applications.
    //
    // Exclusive access is required in order to perform driver escapes.
    //
    hr = g_pJoystick->SetCooperativeLevel( hDlg, 
                                        DISCL_BACKGROUND | DISCL_EXCLUSIVE);
    if ( FAILED(hr) ) 
    {
        g_pJoystick = NULL;
        return hr;
    }

    // Acquire the joystick
    hr = g_pJoystick->Acquire();
    if ( FAILED(hr) ) 
    {
        g_pJoystick = NULL;
        return hr;
    }

    g_bAcquired = TRUE;
    g_nCurrentDevice = n;

    return S_OK;
}


//-----------------------------------------------------------------------------
// Function: ReleaseJoystick
//
// Description: 
//      Unacquire the joystick we are using
//
//-----------------------------------------------------------------------------
HRESULT ReleaseJoystick( HWND hWnd)
{
    if (g_pJoystick)
    {
        g_pJoystick->Unacquire();
        g_pJoystick = NULL;
        g_nCurrentDevice = -1;
        g_bAcquired = FALSE;
    }

    return S_OK;
}


//-----------------------------------------------------------------------------
// Function: GetJoystickState
//
// Description: 
//      Get the current joystick's axis and button states
//
//-----------------------------------------------------------------------------
HRESULT GetJoystickState(DIJOYSTATE *state)
{
    HRESULT     hr;
    DWORD       dwInput = 0;

    // Poll the joystick to read the current state (not really necessary for USB devices)
    hr = g_pJoystick->Poll();

    // Read the device state
    hr = g_pJoystick->GetDeviceState( sizeof(DIJOYSTATE), state );

    return hr;
}


//-----------------------------------------------------------------------------
// Function: GetButtonName
//
// Description: 
//      Get the name of a button on the current joystick
//
//-----------------------------------------------------------------------------
HRESULT GetButtonName(int n, char *name, int bufsiz)
{
    DIDEVICEOBJECTINSTANCE oi;
    HRESULT hr;

    oi.dwSize = sizeof(oi);

    hr = g_pJoystick->GetObjectInfo(&oi, DIJOFS_BUTTON(n), DIPH_BYOFFSET);
    if( FAILED(hr) )
        return E_FAIL;

    strncpy(name, oi.tszName, bufsiz);

    return hr;
}

