//-----------------------------------------------------------------------------
// File: WinMain.cpp
//
// Desc: Windows management for DirectInput sample
//
//
// Copyright (c) 1998 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

#include <windowsx.h>
#include <windows.h>
#include <mmsystem.h>
#include <dinput.h>
#include <commctrl.h>
#include <dbt.h>
#include <stdio.h>

#include "dxsample.h"
#include "resource.h"

#define FEEDBACK_WINDOW_X       220
#define FEEDBACK_WINDOW_Y       50
#define FEEDBACK_WINDOW_WIDTH   160
#define CURSOR_SIZE             3

//-----------------------------------------------------------------------------
// Function prototypes 
//-----------------------------------------------------------------------------
LRESULT CALLBACK MainDialogProc( HWND, UINT, WPARAM, LPARAM );
LRESULT CALLBACK PakDialogProc( HWND, UINT, WPARAM, LPARAM );
void OnPaint( HWND hWnd );
void OnTimer( HWND hWnd );
void InitializeControls( HWND hWnd);
void GetDeviceSettings( HWND hWnd);
void UpdateDeviceList( HWND hWnd);
void SelectJoystick( HWND hWnd);
void SendDirectCommand( HWND hWnd);

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
LONG xPos;
LONG yPos;
BYTE buttons[32];
DIJOYSTATE newState;
BOOL bPosValid = FALSE;

//-----------------------------------------------------------------------------
// Function: WinMain(HANDLE, HANDLE, LPSTR, int)
//
// Description: 
//     Entry point for the application.  Since we use a simple dialog for 
//     user interaction we don't need to pump messages.
//
//-----------------------------------------------------------------------------
int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, 
                      LPSTR lpCmdLine, int nCmdShow )
{
    // DirectInputCreate needs the instance handle
    g_hInst = hInstance;

    InitCommonControls();

    // Display the main dialog box.
    DialogBox( hInstance, 
            MAKEINTRESOURCE(IDD_JOY_FEEDBACK), 
            NULL, 
            (DLGPROC) MainDialogProc );

    FreeDirectInput();

    return TRUE;
}




//-----------------------------------------------------------------------------
// Function: MainDialogProc
//
// Description: 
//      Handles dialog messages
//
//-----------------------------------------------------------------------------
LRESULT CALLBACK MainDialogProc( HWND hDlg, UINT message, WPARAM wParam, 
                                 LPARAM lParam )
{
    HRESULT hr;
    char buf[10];
    int val;
    
    switch (message) 
    {
        case WM_INITDIALOG:
            InitializeControls(hDlg);

            hr = InitDirectInput( hDlg );
            if ( FAILED(hr) )
            {
                MessageBox( NULL, 
                    "Error Initializing DirectInput", 
                    "DirectInput Sample", 
                    MB_ICONERROR | MB_OK );

                EndDialog( hDlg, 0 );
                break;
            }

            UpdateDeviceList(hDlg);

            // Start the timer
            SetTimer(hDlg, 123, 5, NULL);   // 5 millisecond timer (200 polls per second)

            return TRUE;
            break;

        case WM_DEVICECHANGE:
            switch (wParam)
            {
            case DBT_DEVICEARRIVAL:
            case DBT_DEVICEREMOVECOMPLETE:
                UpdateDeviceList(hDlg);
                break;
            }
            break;
        case WM_HSCROLL:
            val = SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
            wsprintf(buf,"%d", val);
        
            if ((HWND)lParam == GetDlgItem( hDlg, IDC_CLIPSLIDER))
            {
                SetDlgItemText(hDlg, IDC_CLIPVALUE, buf);
                SetRangeClipping(val);
            }
            else
            {
                SetDlgItemText(hDlg, IDC_CORNERVALUE, buf);
                SetCornerStretch(val);
            }

            break;

        case WM_TIMER:
            OnTimer( hDlg);
            break;

        case WM_PAINT:
            OnPaint( hDlg );
            break;

        case WM_COMMAND:
            {
                switch ( LOWORD(wParam) )
                {
                case IDC_CLOSE:
                    PostQuitMessage( 0 );
                    break;
                case IDC_RUMBLEGO:
                    if (IsRumbleCommand())
                        SendRumbleCommand(1);
                    break;
                case IDC_RUMBLESTOP:
                    if (IsRumbleCommand())
                        SendRumbleCommand(0);
                    break;
                case IDC_PAKIO:
                    // Display the Pak I/O dialog box.
                    DialogBox( g_hInst, 
                            MAKEINTRESOURCE(IDD_PAK_IO), 
                            NULL, 
                            (DLGPROC) PakDialogProc );
                    break;

                case IDC_DIRECT:
                    // Send a direct command to the controller
                    SendDirectCommand(hDlg);
                    break;

                case IDC_DEVICELIST:
                    if (HIWORD(wParam) == CBN_SELCHANGE)
                    {
                        // User selected a different joystick from the dropdown list
                        SelectJoystick(hDlg);
                    }
                    break;
                }

                return TRUE;
            }
            break;


        case WM_CLOSE:
            EndDialog( hDlg, TRUE ); 

            return TRUE;
            break;
    }

    return FALSE;
}



//-----------------------------------------------------------------------------
// Function: OnPaint
//
// Description: 
//      Handles the WM_PAINT window message
//
//-----------------------------------------------------------------------------
void OnPaint( HWND hWnd )
{
    PAINTSTRUCT ps;
    HDC         hDC;
    HPEN        hpenOld;
    HPEN        hpenBlack;
    HBRUSH      hbrOld;
    HBRUSH      hbrBlack;
    int         x;
    int         y;
    
    hDC = BeginPaint( hWnd, &ps );

    if (NULL != hDC) 
    {
        // everything is scaled to the size of the window.
        hpenBlack = GetStockPen( BLACK_PEN );
        hpenOld = SelectPen( hDC, hpenBlack );

        // draw force feedback bounding rect
        MoveToEx( hDC, FEEDBACK_WINDOW_X - CURSOR_SIZE, 
                       FEEDBACK_WINDOW_Y - CURSOR_SIZE, NULL );

        LineTo( hDC, FEEDBACK_WINDOW_X - CURSOR_SIZE, 
                     FEEDBACK_WINDOW_Y + CURSOR_SIZE + FEEDBACK_WINDOW_WIDTH );

        LineTo( hDC, FEEDBACK_WINDOW_X + CURSOR_SIZE + FEEDBACK_WINDOW_WIDTH, 
                     FEEDBACK_WINDOW_Y + CURSOR_SIZE + FEEDBACK_WINDOW_WIDTH );

        LineTo( hDC, FEEDBACK_WINDOW_X + CURSOR_SIZE + FEEDBACK_WINDOW_WIDTH, 
                     FEEDBACK_WINDOW_Y - CURSOR_SIZE );

        LineTo( hDC, FEEDBACK_WINDOW_X - CURSOR_SIZE, 
                     FEEDBACK_WINDOW_Y - CURSOR_SIZE );

        hbrBlack = GetStockBrush( BLACK_BRUSH );
        hbrOld = SelectBrush( hDC, hbrBlack );

        x = MulDiv( FEEDBACK_WINDOW_WIDTH,
                    xPos - g_xMin, 
                    g_xMax - g_xMin);

        y = MulDiv( FEEDBACK_WINDOW_WIDTH, 
                    yPos - g_yMin, 
                    g_yMax - g_yMin);

        x += FEEDBACK_WINDOW_X;
        y += FEEDBACK_WINDOW_Y;

        if (bPosValid)
            Ellipse( hDC, x-CURSOR_SIZE, y-CURSOR_SIZE, x+CURSOR_SIZE+1, y+CURSOR_SIZE+1 );

        SelectBrush( hDC, hbrOld );
        SelectPen( hDC, hpenOld );

        EndPaint( hWnd, &ps );
    }
}





//-----------------------------------------------------------------------------
// Function: OnTimer
//
void OnTimer( HWND hWnd)
{
    HRESULT hr;
    BOOL bAxisChanged = FALSE;
    BOOL bButtonsChanged = FALSE;
    int i;
    RECT r;

    if (!g_bAcquired)
        return;

    // Update the joystick axis positions and buttons

    hr = GetJoystickState( &newState);
    if( FAILED(hr) )
    {
        return;
    }

    if (!bPosValid)
    {
        bAxisChanged = TRUE;
        bButtonsChanged = TRUE;
    }

    if (newState.lX != xPos || newState.lY != yPos)
    {
        bAxisChanged = TRUE;
    }

    for (i=0; i < g_nButtonCount; i++)
    {
        if ((newState.rgbButtons[i] & 0x80) != buttons[i])
            bButtonsChanged = TRUE;
    }

    bPosValid = TRUE;

    if (bButtonsChanged)
    {
        char butts[200];

        for (i=0; i < g_nButtonCount; i++)
        {
            buttons[i] = newState.rgbButtons[i] & 0x80;
        }

        butts[0] = 0;
        for (i=0; i < g_nButtonCount; i++)
        {
            HRESULT  hr;
            char name[100];

            if (buttons[i] & 0x80)
            {
                hr = GetButtonName(i, name, sizeof(name));
                wsprintf(butts + strlen(butts), "%s  ", name);
            }
        }
        SetDlgItemText( hWnd, IDC_BUTTONS, butts);
    }

    if (bAxisChanged)
    {
        char buf[10];

        xPos = newState.lX;
        yPos = newState.lY;

        wsprintf( buf, "%d", xPos);
        SetDlgItemText( hWnd, IDC_XPOS, buf);

        wsprintf( buf, "%d", yPos);
        SetDlgItemText( hWnd, IDC_YPOS, buf);

        r.left = FEEDBACK_WINDOW_X - CURSOR_SIZE +1;
        r.right = FEEDBACK_WINDOW_X + CURSOR_SIZE+ FEEDBACK_WINDOW_WIDTH;
        r.top = FEEDBACK_WINDOW_Y - CURSOR_SIZE +1;
        r.bottom = FEEDBACK_WINDOW_Y + CURSOR_SIZE+ FEEDBACK_WINDOW_WIDTH;

        InvalidateRect( hWnd, &r, TRUE);
        UpdateWindow( hWnd );
    }
}

void UpdateDeviceList(HWND hWnd)
{
    GUID oldDevice;
    BOOL bOld = FALSE;
    int i;
    HWND hListBox = GetDlgItem(hWnd, IDC_DEVICELIST);

    // remember who we had selected
    if (g_bAcquired)
    {
        oldDevice = g_gDeviceGuids[g_nCurrentDevice];
        bOld = TRUE;
    }

    EnumerateJoysticks( hWnd );

    // Fill list box
    SendMessage( hListBox, CB_RESETCONTENT, 0, 0);
    for (i=0; i < g_nDevices; i++)
    {
        SendMessage( hListBox, CB_ADDSTRING, 0, (LPARAM) g_pDeviceNames[i]);
    }

    // Try to re-select the we we had previously selected
    i = 0;
    if (bOld)
    {
        for (i=0; i < g_nDevices; i++)
        {
            if (oldDevice == g_gDeviceGuids[i])
            {
                break;
            }
        }

        // if we didn't find it, use the first one
        if (i == g_nDevices)
            i = 0;
    }

    SendMessage( hListBox, CB_SETCURSEL, i, 0);
    SelectJoystick( hWnd);
}

void InitializeControls(HWND hWnd)
{
    // Range Clipping Slider
    HWND hClipSlider = GetDlgItem( hWnd, IDC_CLIPSLIDER);
    SendMessage( hClipSlider, TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(50, 130));
    SetDlgItemText(hWnd, IDC_CLIPVALUE, "");

    // Corner Stretch Slider
    HWND hCornerSlider = GetDlgItem( hWnd, IDC_CORNERSLIDER);
    SendMessage( hCornerSlider, TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(0, 20));
    SetDlgItemText(hWnd, IDC_CORNERVALUE, "");
}

void SelectJoystick(HWND hWnd)
{
    HRESULT hr;
    int n;
    HWND hListBox = GetDlgItem(hWnd, IDC_DEVICELIST);

    // unselect old joystick
    ReleaseJoystick( hWnd);

    // select new one
    n = SendMessage( hListBox, CB_GETCURSEL, 0, 0);
    if (n != CB_ERR)
    {
        hr = UseJoystick( hWnd, n );
    }

    GetDeviceSettings(hWnd);
}

void GetDeviceSettings(HWND hWnd)
{
    char devinfobuf[50];
    int dnlen = 0;
    DWORD dnval;
    HRESULT hr;

    HWND hClipLabel  = GetDlgItem( hWnd, IDC_CLIPLABEL);
    HWND hClipValue  = GetDlgItem( hWnd, IDC_CLIPVALUE);
    HWND hClipSlider = GetDlgItem( hWnd, IDC_CLIPSLIDER);
    HWND hClipLo     = GetDlgItem( hWnd, IDC_CLIPLO);
    HWND hClipHi     = GetDlgItem( hWnd, IDC_CLIPHI);

    HWND hCornerLabel  = GetDlgItem( hWnd, IDC_CORNERLABEL);
    HWND hCornerValue  = GetDlgItem( hWnd, IDC_CORNERVALUE);
    HWND hCornerSlider = GetDlgItem( hWnd, IDC_CORNERSLIDER);
    HWND hCornerLo     = GetDlgItem( hWnd, IDC_CORNERLO);
    HWND hCornerHi     = GetDlgItem( hWnd, IDC_CORNERHI);

    HWND hRumbleGo     = GetDlgItem( hWnd, IDC_RUMBLEGO);
    HWND hRumbleStop   = GetDlgItem( hWnd, IDC_RUMBLESTOP);
    HWND hPakIO        = GetDlgItem( hWnd, IDC_PAKIO);
    HWND hDirect       = GetDlgItem( hWnd, IDC_DIRECT);

    HWND hDevInfo      = GetDlgItem( hWnd, IDC_DEVINFO);

    EnableWindow( hClipLabel, FALSE);
    EnableWindow( hClipValue, FALSE);
    EnableWindow( hClipSlider, FALSE);
    EnableWindow( hClipLo, FALSE);
    EnableWindow( hClipHi, FALSE);

    EnableWindow( hCornerLabel, FALSE);
    EnableWindow( hCornerValue, FALSE);
    EnableWindow( hCornerSlider, FALSE);
    EnableWindow( hCornerLo, FALSE);
    EnableWindow( hCornerHi, FALSE);

    EnableWindow( hRumbleGo, FALSE);
    EnableWindow( hRumbleStop, FALSE);
    EnableWindow( hPakIO, FALSE);
    EnableWindow( hDirect, FALSE);

    EnableWindow( hDevInfo, FALSE);

    SetDlgItemText( hWnd, IDC_DEVINFO, "N/A");

    SetDlgItemText( hWnd, IDC_BUTTONCOUNT, "");

    if (!g_bAcquired)
        return;

    // Show number of buttons
    char buf[10];
    wsprintf(buf, "%d", g_nButtonCount);
    SetDlgItemText( hWnd, IDC_BUTTONCOUNT, buf);

    // Device number & port name
    devinfobuf[0] = 0;
    if (IsDeviceNumber())
    {
        hr = GetDeviceNumber( &dnval);
        if (!FAILED(hr))
        {
            wsprintf(devinfobuf,"Device #%d", dnval);
            dnlen = strlen(devinfobuf);
        }
    }
    
    if (IsPortName())
    {
        char portbuf[40];
        hr = GetPortName( portbuf, sizeof(portbuf));
        if (!FAILED(hr))
        {
            wsprintf(devinfobuf+dnlen,", port %s", portbuf);
        }
    }

    if (devinfobuf[0] != 0)
    {
        SetDlgItemText(hWnd, IDC_DEVINFO, devinfobuf);
        EnableWindow( hDevInfo, TRUE);
    }

    // Rumble Command
    if (IsRumbleCommand())
    {
        EnableWindow( hRumbleGo, TRUE);
        EnableWindow( hRumbleStop, TRUE);
    }

    // Rumble Command
    if (IsPakIO())
    {
        EnableWindow( hPakIO, TRUE);
    }

    // Direct Command
    if (IsDirectCommand())
    {
        EnableWindow( hDirect, TRUE);
    }

    // Range Clipping Slider
    if (IsRangeClipping())
    {
        DWORD val;

        hr = GetRangeClipping( &val);
        if (!FAILED(hr))
        {
            char buf[20];

            EnableWindow( hClipLabel, TRUE);
            EnableWindow( hClipValue, TRUE);
            EnableWindow( hClipSlider, TRUE);
            EnableWindow( hClipLo, TRUE);
            EnableWindow( hClipHi, TRUE);
 
            wsprintf(buf,"%d", val);
            SetDlgItemText(hWnd, IDC_CLIPVALUE, buf);

            SendMessage( hClipSlider, TBM_SETPOS, TRUE, val);
        }
    }

    // Corner Stretch Slider
    if (IsCornerStretch())
    {
        DWORD val;

        hr = GetCornerStretch( &val);
        if (!FAILED(hr))
        {
            char buf[20];

            EnableWindow( hCornerLabel, TRUE);
            EnableWindow( hCornerValue, TRUE);
            EnableWindow( hCornerSlider, TRUE);
            EnableWindow( hCornerLo, TRUE);
            EnableWindow( hCornerHi, TRUE);
 
            wsprintf(buf,"%d", val);
            SetDlgItemText(hWnd, IDC_CORNERVALUE, buf);

            SendMessage( hCornerSlider, TBM_SETPOS, TRUE, val);
        }
    }
}


void SendDirectCommand( HWND hWnd)
 {
    HRESULT hr;
    BYTE to[1];
    BYTE from[4];
    int fromcount;
    int i;
    char buf[100];

    to[0] = 1;       // for this example we'll send the "1" command
    fromcount = 4;   // expecting 4 bytes back

    hr = DirectCommand( to, 1, from, &fromcount);
    if (SUCCEEDED(hr))
    {
        wsprintf(buf, "Controller command 1 returned %d bytes:\n\n", fromcount);
        for (i=0; i < fromcount; i++)
        {
            wsprintf(buf + strlen(buf), "%02x ", from[i]);
        }
    }
    else
    {
        wsprintf(buf, "Controller command 1 failed");
    }
    MessageBox( hWnd, buf, "Direct Command Test", MB_OK);
 }


 //-----------------------------------------------------------------------------
// Function: PakDialogProc
//
// Description: 
//      Handles dialog messages
//
//-----------------------------------------------------------------------------
LRESULT CALLBACK PakDialogProc( HWND hDlg, UINT message, WPARAM wParam, 
                                 LPARAM lParam )
{
    char buf[100];
    HRESULT hr;
    FILE *fp;
    int i;

    switch (message) 
    {
        case WM_COMMAND:
            {
                switch ( LOWORD(wParam) )
                {
                case IDC_PAKINIT:
                    BYTE ret;
                    hr = InitializePak( &ret);
                    if (SUCCEEDED(hr))
                    {
                        wsprintf(buf, "Pak Initialization returned 0x%02X", ret);
                    }
                    else
                    {
                        wsprintf(buf, "Pak Initialization failed");
                    }

                    SetDlgItemText( hDlg, IDC_PAKIORESULT, buf);
                    break;

                case IDC_PAKREAD:
                    fp = fopen("dxsample.pak","wb");
                    if (fp)
                    {
                        HCURSOR oldCursor = SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_WAIT)));

                        for (i=0; i < 32768; i += 32)
                        {
                            BYTE data[32];

                            hr = ReadPak( i, data);
                            if (FAILED(hr))
                            {
                                wsprintf(buf, "Pak Read: error reading from Pak, addr=%04X", i);
                                break;
                            }

                            fwrite( data, 32, 1, fp);

                            wsprintf(buf, "Pak Read: completed %dK of 32K bytes", (i+32)/1024);
                            SetDlgItemText( hDlg, IDC_PAKIORESULT, buf);
                            UpdateWindow(hDlg);
                        }
                        fclose(fp);

                        SetCursor(oldCursor);
                    }
                    else
                    {
                        wsprintf(buf, "Pak Read: could not create dxsample.pak");
                    }

                    SetDlgItemText( hDlg, IDC_PAKIORESULT, buf);
                    break;

                case IDC_PAKWRITE:
                    fp = fopen("dxsample.pak","rb");
                    if (fp)
                    {
                        HCURSOR oldCursor = SetCursor(LoadCursor(NULL,MAKEINTRESOURCE(IDC_WAIT)));
    
                        for (i=0; i < 32768; i += 32)
                        {
                            BYTE data[32];

                            fread( data, 32, 1, fp);

                            hr = WritePak( i, data);
                            if (FAILED(hr))
                            {
                                wsprintf(buf, "Pak Read: error writing to Pak, addr=%04X", i);
                                break;
                            }

                            wsprintf(buf, "Pak Write: completed %dK of 32K bytes", (i+32)/1024);
                            SetDlgItemText( hDlg, IDC_PAKIORESULT, buf);
                            UpdateWindow(hDlg);
                        }
                        fclose(fp);

                        SetCursor(oldCursor);
                    }
                    else
                    {
                        wsprintf(buf, "Pak Write: could not open dxsample.pak");
                    }

                    SetDlgItemText( hDlg, IDC_PAKIORESULT, buf);
                    break;

                case IDOK:
                    EndDialog( hDlg, TRUE ); 
                    break;

                }

                return TRUE;
            }
            break;


        case WM_CLOSE:
            EndDialog( hDlg, TRUE ); 

            return TRUE;
    }

    return FALSE;
}



